from .arduino import AnalogPins, GPIOPinMode
from .astoria import RobotMode
from .exceptions import (
    BoardDisconnectionError, MetadataKeyError, MetadataNotReadyError,
)
from .kch import Colour, UserLED
from .logging import add_trace_level
from .motor_board import MotorPower
from .power_board import Note, PowerOutputPosition
from .raw_serial import RawSerialDevice
from .robot import Robot
from .utils import list_ports

add_trace_level()

BRAKE = MotorPower.BRAKE
COAST = MotorPower.COAST

A0 = AnalogPins.A0
A1 = AnalogPins.A1
A2 = AnalogPins.A2
A3 = AnalogPins.A3
A4 = AnalogPins.A4
A5 = AnalogPins.A5

COMP = RobotMode.COMP
DEV = RobotMode.DEV

INPUT = GPIOPinMode.INPUT
INPUT_PULLUP = GPIOPinMode.INPUT_PULLUP
OUTPUT = GPIOPinMode.OUTPUT

OUT_H0 = PowerOutputPosition.H0
OUT_H1 = PowerOutputPosition.H1
OUT_L0 = PowerOutputPosition.L0
OUT_L1 = PowerOutputPosition.L1
OUT_L3 = PowerOutputPosition.L3
OUT_FIVE_VOLT = PowerOutputPosition.FIVE_VOLT

LED_A = UserLED.A
LED_B = UserLED.B
LED_C = UserLED.C

__all__ = [
    "A0",
    "A1",
    "A2",
    "A3",
    "A4",
    "A5",
    "BRAKE",
    "BoardDisconnectionError",
    "COAST",
    "COMP",
    "Colour",
    "DEV",
    "INPUT",
    "INPUT_PULLUP",
    "LED_A",
    "LED_B",
    "LED_C",
    "MetadataKeyError",
    "MetadataNotReadyError",
    "Note",
    "OUTPUT",
    "OUT_FIVE_VOLT",
    "OUT_H0",
    "OUT_H1",
    "OUT_L0",
    "OUT_L1",
    "OUT_L3",
    "RawSerialDevice",
    "Robot",
    "list_ports",
]



"""Test that the module works."""
import logging
import signal
import socket

import pytest

from sr.robot3 import Robot
from sr.robot3.exceptions import MetadataNotReadyError
from sr.robot3.utils import BoardIdentity, obtain_lock

from .conftest import MockAtExit, MockSerialWrapper


def test_robot(monkeypatch, caplog) -> None:
    """Test that the Robot object can be created."""
    # monkey patch serial ports so we can test without hardware
    monkeypatch.setattr('sr.robot3.power_board.SerialWrapper', MockSerialWrapper([
        ("*IDN?", "Student Robotics:PBv4B:POW123:4.4.1"),
        ("OUT:0:SET:1", "ACK"),
        ("OUT:1:SET:1", "ACK"),
        ("OUT:2:SET:1", "ACK"),
        ("OUT:3:SET:1", "ACK"),
        ("OUT:5:SET:1", "ACK"),
        ("OUT:6:SET:1", "ACK"),
        ("*IDN?", "Student Robotics:PBv4B:POW123:4.4.1"),
        ("BTN:START:GET?", "0:1"),
        ("NOTE:1760:100", "ACK"),  # Start up sound
        ("LED:RUN:SET:F", "ACK"),
        ("BTN:START:GET?", "0:0"),
        ("BTN:START:GET?", "0:1"),
        ("LED:RUN:SET:1", "ACK"),
    ]))
    monkeypatch.setattr('sr.robot3.motor_board.SerialWrapper', MockSerialWrapper([
        ("*IDN?", "Student Robotics:MCv4B:MOT123:4.4"),
        ("*IDN?", "Student Robotics:MCv4B:MOT123:4.4"),
    ]))
    monkeypatch.setattr('sr.robot3.servo_board.SerialWrapper', MockSerialWrapper([
        ("*IDN?", "Student Robotics:SBv4B:TEST123:4.3"),
        ("*IDN?", "Student Robotics:SBv4B:TEST123:4.3"),
    ]))
    monkeypatch.setattr('sr.robot3.arduino.SerialWrapper', MockSerialWrapper([
        ("v", "SRduino:2.0"),
        ("v", "SRduino:2.0"),
    ]))

    # monkey patch atexit to avoid running cleanup code
    monkeypatch.setattr('sr.robot3.power_board.atexit', MockAtExit())
    monkeypatch.setattr('sr.robot3.motor_board.atexit', MockAtExit())
    monkeypatch.setattr('sr.robot3.servo_board.atexit', MockAtExit())

    # Monkey patch serial comport lookup so only manual boards are found
    monkeypatch.setattr('sr.robot3.power_board.comports', lambda: [])
    monkeypatch.setattr('sr.robot3.motor_board.comports', lambda: [])
    monkeypatch.setattr('sr.robot3.servo_board.comports', lambda: [])
    monkeypatch.setattr('sr.robot3.arduino.comports', lambda: [])

    # Forget the camera
    monkeypatch.setattr('sr.robot3.robot._setup_cameras', lambda *_: {})

    manual_boards = {
        'PBv4B': ['test://'],
        'SBv4B': ['test://'],
        'MCv4B': ['test://'],
        'Arduino': ['test://'],
    }
    # check logging
    caplog.clear()
    caplog.set_level(logging.CRITICAL, logger='sr.robot3.mqtt')
    caplog.set_level(logging.CRITICAL, logger='sr.robot3.astoria')
    caplog.set_level(logging.INFO, logger='sr.robot3.robot')

    # Test that we can obtain a lock before creating a robot object
    lock = obtain_lock()
    assert isinstance(lock, socket.socket)
    lock.close()  # release the lock

    # Test that we can create a robot object
    r = Robot(wait_for_start=False, manual_boards=manual_boards, debug=True)
    assert caplog.record_tuples[1:] == [
        # First line contains the version number
        ('sr.robot3.robot', logging.INFO, 'Found PowerBoard, serial: POW123'),
        ('sr.robot3.robot', logging.INFO, 'Found MotorBoard, serial: MOT123'),
        ('sr.robot3.robot', logging.INFO, 'Found ServoBoard, serial: TEST123'),
        ('sr.robot3.robot', logging.INFO, 'Found Arduino, serial: test://'),
    ]

    # Check we found all the boards
    assert r.power_board._identity == BoardIdentity(
        "Student Robotics", "PBv4B", "POW123", "4.4.1")

    assert r.motor_board._identity == BoardIdentity(
        "Student Robotics", "MCv4B", "MOT123", "4.4")

    assert r.servo_board._identity == BoardIdentity(
        "Student Robotics", "SBv4B", "TEST123", "4.3")

    assert r.arduino._identity == BoardIdentity(
        "Student Robotics", "SRduino", "test://", "2.0")

    # Check that a RuntimeError is raised if we have 0 instances of a board
    with pytest.raises(RuntimeError, match="No boards of this type found"):
        r.camera.identify()

    # Check that we can't obtain a lock when a robot object already exists
    with pytest.raises(
        OSError,
        match="Unable to obtain lock, Is another robot instance already running?"
    ):
        obtain_lock()

    # Check that an exception is raised if we try to access metadata before wait_start
    with pytest.raises(
        MetadataNotReadyError,
        match=".*can only be used after wait_start has been called"
    ):
        r.zone

    r.wait_start()

    assert r.zone == 0

    # Check _something_ is configured to handle SIGTERM
    assert signal.getsignal(signal.SIGTERM) != signal.SIG_DFL


@pytest.hookimpl(trylast=True)
@pytest.mark.hardware
def test_robot_discovery() -> None:
    """Test that we can discover all the boards when creating a Robot object."""
    robot = Robot(wait_for_start=False)

    # Test that we can access the singular boards
    power_asset_tag = robot.power_board.identify().asset_tag
    servo_asset_tag = robot.servo_board.identify().asset_tag
    motor_asset_tag = robot.motor_board.identify().asset_tag

    # Test that we can access the boards by asset tag
    assert robot.power_board == robot.boards[power_asset_tag]
    assert robot.servo_board == robot.boards[servo_asset_tag]
    assert robot.motor_board == robot.boards[motor_asset_tag]


def test_cannot_subclass() -> None:
    with pytest.raises(TypeError, match="Robot cannot be sub-classed"):
        class MyRobot(Robot):
            pass



"""Game specific code."""
from __future__ import annotations

from typing import Iterable

# Marker sizes are in mm
MARKER_SIZES: dict[Iterable[int], int] = {
    range(50): 150,  # 0-49 = 150mm
    range(50, 100): 200,  # 50-99 = 200mm
    range(100, 200): 80,  # 100-199 = 80mm
}



"""The servo board module provides an interface to the servo board firmware over serial."""
from __future__ import annotations

import atexit
import logging
from types import MappingProxyType
from typing import NamedTuple, Optional

from serial.tools.list_ports import comports

from .exceptions import IncorrectBoardError
from .logging import log_to_debug
from .serial_wrapper import SerialWrapper
from .utils import (
    IN_SIMULATOR, Board, BoardIdentity, float_bounds_check,
    get_simulator_boards, get_USB_identity, map_to_float, map_to_int,
)

DUTY_MIN = 500
DUTY_MAX = 4000
START_DUTY_MIN = 1000
START_DUTY_MAX = 2000
NUM_SERVOS = 12

logger = logging.getLogger(__name__)
BAUDRATE = 115200  # Since the servo board is a USB device, this is ignored


class ServoStatus(NamedTuple):
    """A named tuple containing the values of the servo status output."""
    watchdog_failed: bool
    power_good: bool

    @classmethod
    def from_status_response(cls, response: str) -> ServoStatus:
        """
        Create a ServoStatus from a status response.

        :param response: The response from a *STATUS? command.
        :return: The ServoStatus.
        """
        data = response.split(':')

        return cls(
            watchdog_failed=(data[0] == '1'),
            power_good=(data[1] == '1'),
        )


class ServoBoard(Board):
    """
    A class representing the servo board interface.

    This class is intended to be used to communicate with the servo board over serial
    using the text-based protocol added in version 4.3 of the servo board firmware.

    :param serial_port: The serial port to connect to.
    :param initial_identity: The identity of the board, as reported by the USB descriptor.
    """
    __slots__ = ('_serial', '_identity', '_servos')

    @staticmethod
    def get_board_type() -> str:
        """
        Return the type of the board.

        :return: The literal string 'SBv4B'.
        """
        return 'SBv4B'

    def __init__(
        self,
        serial_port: str,
        initial_identity: Optional[BoardIdentity] = None,
    ) -> None:
        if initial_identity is None:
            initial_identity = BoardIdentity()
        self._serial = SerialWrapper(serial_port, BAUDRATE, identity=initial_identity)

        self._servos = tuple(
            Servo(self._serial, index) for index in range(NUM_SERVOS)
        )

        self._identity = self.identify()
        if self._identity.board_type != self.get_board_type():
            raise IncorrectBoardError(self._identity.board_type, self.get_board_type())
        self._serial.set_identity(self._identity)

        atexit.register(self._cleanup)

    @classmethod
    def _get_valid_board(
        cls,
        serial_port: str,
        initial_identity: Optional[BoardIdentity] = None,
    ) -> Optional[ServoBoard]:
        """
        Attempt to connect to a servo board and returning None if it fails identification.

        :param serial_port: The serial port to connect to.
        :param initial_identity: The identity of the board, as reported by the USB descriptor.

        :return: A ServoBoard object, or None if the board could not be identified.
        """
        try:
            board = cls(serial_port, initial_identity)
        except IncorrectBoardError as err:
            logger.warning(
                f"Board returned type {err.returned_type!r}, "
                f"expected {err.expected_type!r}. Ignoring this device")
            return None
        except Exception:
            if initial_identity is not None:
                if initial_identity.board_type == 'manual':
                    logger.warning(
                        f"Manually specified servo board at port {serial_port!r}, "
                        "could not be identified. Ignoring this device")
                elif initial_identity.manufacturer == 'sbot_simulator':
                    logger.warning(
                        f"Simulator specified servo board at port {serial_port!r}, "
                        "could not be identified. Ignoring this device")
                return None

            logger.warning(
                f"Found servo board-like serial port at {serial_port!r}, "
                "but it could not be identified. Ignoring this device")
            return None
        return board

    @classmethod
    def _get_simulator_boards(cls) -> MappingProxyType[str, ServoBoard]:
        """
        Get the simulator boards.

        :return: A mapping of board serial numbers to boards.
        """
        boards = {}
        # The filter here is the name of the emulated board in the simulator
        for board_info in get_simulator_boards('ServoBoard'):

            # Create board identity from the info given
            initial_identity = BoardIdentity(
                manufacturer='sbot_simulator',
                board_type=board_info.type_str,
                asset_tag=board_info.serial_number,
            )
            if (board := cls._get_valid_board(board_info.url, initial_identity)) is None:
                continue

            boards[board._identity.asset_tag] = board
        return MappingProxyType(boards)

    @classmethod
    def _get_supported_boards(
        cls, manual_boards: Optional[list[str]] = None,
    ) -> MappingProxyType[str, 'ServoBoard']:
        """
        Find all connected servo boards.

        Ports are filtered to the USB vendor and product ID: 0x1BDA and 0x0011 respectively.

        :param manual_boards: A list of manually specified serial ports to also attempt
            to connect to, defaults to None
        :return: A mapping of serial numbers to servo boards.
        """
        if IN_SIMULATOR:
            return cls._get_simulator_boards()

        boards = {}
        serial_ports = comports()
        for port in serial_ports:
            # Filter to USB vendor and product ID of the SR v4 servo board
            if port.vid == 0x1BDA and port.pid == 0x0011:
                # Create board identity from USB port info
                initial_identity = get_USB_identity(port)

                if (board := cls._get_valid_board(port.device, initial_identity)) is None:
                    continue

                boards[board._identity.asset_tag] = board

        # Add any manually specified boards
        if isinstance(manual_boards, list):
            for manual_port in manual_boards:
                # Create board identity from the info given
                initial_identity = BoardIdentity(
                    board_type='manual',
                    asset_tag=manual_port,
                )

                if (board := cls._get_valid_board(manual_port, initial_identity)) is None:
                    continue

                boards[board._identity.asset_tag] = board
        return MappingProxyType(boards)

    @property
    @log_to_debug
    def servos(self) -> tuple['Servo', ...]:
        """
        A tuple of the servos on the board.

        :return: A tuple of the servos on the board.
        """
        return self._servos

    @log_to_debug
    def identify(self) -> BoardIdentity:
        """
        Get the identity of the board.

        :return: The identity of the board.
        """
        response = self._serial.query('*IDN?')
        return BoardIdentity(*response.split(':'))

    @log_to_debug
    def status(self) -> ServoStatus:
        """
        Get the board's status.

        :return: A named tuple of the watchdog fail and pgood status.
        """
        response = self._serial.query('*STATUS?')

        return ServoStatus.from_status_response(response)

    @log_to_debug
    def reset(self) -> None:
        """
        Reset the board.

        This will disable all servos.
        """
        self._serial.write('*RESET')

    @property
    @log_to_debug
    def current(self) -> float:
        """
        Get the current draw of the board.

        This only includes the servos powered through the main port, not the aux port.

        :return: The current draw of the board in amps.
        """
        response = self._serial.query('SERVO:I?')
        return float(response) / 1000

    @property
    @log_to_debug
    def voltage(self) -> float:
        """
        Get the voltage of the on-board regulator.

        :return: The voltage of the on-board regulator in volts.
        """
        response = self._serial.query('SERVO:V?')
        return float(response) / 1000

    def _cleanup(self) -> None:
        """
        Reset the board and disable all servos on exit.

        This is registered as an exit function.
        """
        try:
            self.reset()
        except Exception:
            logger.warning(f"Failed to cleanup servo board {self._identity.asset_tag}.")

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial}>"


class Servo:
    """
    A class representing a servo on the servo board.

    :param serial: The serial wrapper to use to communicate with the board.
    :param index: The index of the servo on the board.
    """
    __slots__ = ('_serial', '_index', '_duty_min', '_duty_max')

    def __init__(self, serial: SerialWrapper, index: int):
        self._serial = serial
        self._index = index

        self._duty_min = START_DUTY_MIN
        self._duty_max = START_DUTY_MAX

    @log_to_debug
    def set_duty_limits(self, lower: int, upper: int) -> None:
        """
        Set the pulse on-time limits of the servo.

        These limits are used to map the servo position to a pulse on-time.

        :param lower: The lower limit of the servo pulse in µs.
        :param upper: The upper limit of the servo pulse in µs.
        :raises TypeError: If the limits are not ints.
        :raises ValueError: If the limits are not in the range 500 to 4000.
        """
        if not (isinstance(lower, int) and isinstance(upper, int)):
            raise TypeError(
                f'Servo pulse limits are ints in µs, in the range {DUTY_MIN} to {DUTY_MAX}'
            )
        if not (DUTY_MIN <= lower <= DUTY_MAX and DUTY_MIN <= upper <= DUTY_MAX):
            raise ValueError(
                f'Servo pulse limits are ints in µs, in the range {DUTY_MIN} to {DUTY_MAX}'
            )

        self._duty_min = lower
        self._duty_max = upper

    @log_to_debug
    def get_duty_limits(self) -> tuple[int, int]:
        """
        Get the current pulse on-time limits of the servo.

        The limits are specified in µs.

        :return: A tuple of the lower and upper limits of the servo pulse in µs.
        """
        return self._duty_min, self._duty_max

    @property
    @log_to_debug
    def position(self) -> Optional[float]:
        """
        Get the position of the servo.

        If the servo is disabled, this will return None.

        :return: The position of the servo as a float between -1.0 and 1.0 or None if disabled.
        """
        response = self._serial.query(f'SERVO:{self._index}:GET?')
        data = int(response)
        if data == 0:
            return None
        return map_to_float(data, self._duty_min, self._duty_max, -1.0, 1.0, precision=3)

    @position.setter
    @log_to_debug(setter=True)
    def position(self, value: Optional[float]) -> None:
        """
        Set the position of the servo.

        If the servo is disabled, this will enable it.
        -1.0 to 1.0 may not be the full range of the servo, see set_duty_limits().

        :param value: The position of the servo as a float between -1.0 and 1.0
            or None to disable.
        """
        if value is None:
            self.disable()
            return
        value = float_bounds_check(
            value, -1.0, 1.0,
            'Servo position is a float between -1.0 and 1.0')

        setpoint = map_to_int(value, -1.0, 1.0, self._duty_min, self._duty_max)
        self._serial.write(f'SERVO:{self._index}:SET:{setpoint}')

    @log_to_debug
    def disable(self) -> None:
        """
        Disable the servo.

        This will cause this channel to output a 0% duty cycle.
        """
        self._serial.write(f'SERVO:{self._index}:DISABLE')

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__} index={self._index} {self._serial}>"


if __name__ == '__main__':  # pragma: no cover
    servoboards = ServoBoard._get_supported_boards()
    for serial_num, board in servoboards.items():
        print(serial_num)
        print(board.identify())



import csv
import json
from math import radians
from pathlib import Path
from typing import NamedTuple

import pytest
import numpy as np

from sr.robot3.marker import Marker


class MarkerValues(NamedTuple):
    index: int
    tvec: np.ndarray
    rvec: np.ndarray
    distance: float
    horizontal_angle: float
    vertical_angle: float
    yaw: float
    pitch: float
    roll: float
    name: str = "marker"


class MockAprilMarker(NamedTuple):
    """
    A minimal mock of the Marker class from april_vision.
    """
    id: int
    size: int
    tvec: np.ndarray
    rvec: np.ndarray
    pixel_corners: tuple = ((0, 0), (0, 0), (0, 0), (0, 0))
    pixel_centre: tuple = (0, 0)


def load_test_values():
    with open(Path(__file__).parent / "test_data/marker_detections.json") as f:
        test_data = json.load(f)

    expected_values = []
    with open(Path(__file__).parent / "test_data/marker_locations.csv") as f:
        csv_reader = csv.DictReader(f)
        for row in csv_reader:
            expected_values.append(row)

    return [
        pytest.param(
            MarkerValues(
                index=int(row["Index"]),
                tvec=np.array(test_data[row["Index"]]["tvec"]),
                rvec=np.array(test_data[row["Index"]]["rvec"]),
                distance=float(row["distance"]),
                horizontal_angle=float(row["horizontal angle"]),
                vertical_angle=float(row["vertical angle"]),
                yaw=float(row["yaw"]),
                pitch=float(row["pitch"]),
                roll=float(row["roll"]),
                name=row.get("name", "marker"),
            ),
            id=row.get("name", "marker"),
        )
        for row in expected_values
    ]


@pytest.mark.parametrize("test_values", load_test_values())
def test_marker(test_values: MarkerValues):
    marker_under_test = MockAprilMarker(
        id=14, size=200, tvec=test_values.tvec, rvec=test_values.rvec,
    )
    marker = Marker.from_april_vision_marker(marker_under_test)

    assert marker.id == 14, "Marker ID should be 14"
    assert marker.size == 200, "Marker size should be 200mm"

    # in mm
    assert marker.position.distance == pytest.approx(
        test_values.distance * 1000, rel=1e-2), (
        f"Distance to marker of {test_values.name} is incorrect. "
        f"{marker.position.distance} != {test_values.distance * 1000}")
    assert marker.position.horizontal_angle == pytest.approx(
        test_values.horizontal_angle, abs=radians(0.1)), (
        f"Horizontal angle of {test_values.name} is incorrect. "
        f"{marker.position.horizontal_angle} != {test_values.horizontal_angle}")
    assert marker.position.vertical_angle == pytest.approx(
        test_values.vertical_angle, abs=radians(0.1)), (
        f"Vertical angle of {test_values.name} is incorrect. "
        f"{marker.position.vertical_angle} != {test_values.vertical_angle}")

    assert marker.orientation.yaw == pytest.approx(test_values.yaw, abs=radians(2.5)), (
        f"Yaw of {test_values.name} is incorrect. "
        f"{marker.orientation.yaw} != {test_values.yaw}"
    )
    assert marker.orientation.pitch == pytest.approx(test_values.pitch, abs=radians(2.5)), (
        f"Pitch of {test_values.name} is incorrect. "
        f"{marker.orientation.pitch} != {test_values.pitch}"
    )
    assert marker.orientation.roll == pytest.approx(test_values.roll, abs=radians(0.2)), (
        f"Roll of {test_values.name} is incorrect. "
        f"{marker.orientation.roll} != {test_values.roll}"
    )



"""The motor board module provides an interface to the motor board firmware over serial."""
from __future__ import annotations

import atexit
import logging
from enum import IntEnum
from types import MappingProxyType
from typing import NamedTuple, Optional

from serial.tools.list_ports import comports

from .exceptions import IncorrectBoardError
from .logging import log_to_debug
from .serial_wrapper import SerialWrapper
from .utils import (
    IN_SIMULATOR, Board, BoardIdentity, float_bounds_check,
    get_simulator_boards, get_USB_identity, map_to_float, map_to_int,
)

logger = logging.getLogger(__name__)
BAUDRATE = 115200


class MotorPower(IntEnum):
    """Special values for motor power."""
    BRAKE = 0
    COAST = -1024  # A value outside the allowable range


class MotorStatus(NamedTuple):
    """A tuple representing the status of the motor board."""
    output_faults: tuple[bool, ...]
    input_voltage: float
    other: list[str] = []

    @classmethod
    def from_status_response(cls, response: str) -> MotorStatus:
        """
        Create a MotorStatus object from the response to a status command.

        :param response: The response from a *STATUS? command.
        :raise TypeError: If the response is invalid.
        :return: A MotorStatus object.
        """
        output_fault_str, input_voltage_mv, *other = response.split(':')
        output_faults = tuple((port == '1') for port in output_fault_str.split(','))
        input_voltage = float(input_voltage_mv) / 1000
        return cls(output_faults, input_voltage, other)


class MotorBoard(Board):
    """
    A class representing the motor board interface.

    This class is intended to be used to communicate with the motor board over serial
    using the text-based protocol added in version 4.4 of the motor board firmware.

    :param serial_port: The serial port to connect to.
    :param initial_identity: The identity of the board, as reported by the USB descriptor.
    """
    __slots__ = ('_serial', '_identity', '_motors')

    @staticmethod
    def get_board_type() -> str:
        """
        Return the type of the board.

        :return: The literal string 'MCv4B'.
        """
        return 'MCv4B'

    def __init__(
        self,
        serial_port: str,
        initial_identity: Optional[BoardIdentity] = None,
    ) -> None:
        if initial_identity is None:
            initial_identity = BoardIdentity()
        self._serial = SerialWrapper(serial_port, BAUDRATE, identity=initial_identity)

        self._motors = (
            Motor(self._serial, 0),
            Motor(self._serial, 1)
        )

        self._identity = self.identify()
        if self._identity.board_type != self.get_board_type():
            raise IncorrectBoardError(self._identity.board_type, self.get_board_type())
        self._serial.set_identity(self._identity)

        # Disable motors on exit
        atexit.register(self._cleanup)

    @classmethod
    def _get_valid_board(
        cls,
        serial_port: str,
        initial_identity: Optional[BoardIdentity] = None,
    ) -> Optional[MotorBoard]:
        """
        Attempt to connect to a motor board and returning None if it fails identification.

        :param serial_port: The serial port to connect to.
        :param initial_identity: The identity of the board, as reported by the USB descriptor.

        :return: A MotorBoard object, or None if the board could not be identified.
        """
        try:
            board = cls(serial_port, initial_identity)
        except IncorrectBoardError as err:
            logger.warning(
                f"Board returned type {err.returned_type!r}, "
                f"expected {err.expected_type!r}. Ignoring this device")
            return None
        except Exception:
            if initial_identity is not None:
                if initial_identity.board_type == 'manual':
                    logger.warning(
                        f"Manually specified motor board at port {serial_port!r}, "
                        "could not be identified. Ignoring this device")
                elif initial_identity.manufacturer == 'sbot_simulator':
                    logger.warning(
                        f"Simulator specified motor board at port {serial_port!r}, "
                        "could not be identified. Ignoring this device")
                return None

            logger.warning(
                f"Found motor board-like serial port at {serial_port!r}, "
                "but it could not be identified. Ignoring this device")
            return None
        return board

    @classmethod
    def _get_simulator_boards(cls) -> MappingProxyType[str, MotorBoard]:
        """
        Get the simulator boards.

        :return: A mapping of board serial numbers to boards.
        """
        boards = {}
        # The filter here is the name of the emulated board in the simulator
        for board_info in get_simulator_boards('MotorBoard'):

            # Create board identity from the info given
            initial_identity = BoardIdentity(
                manufacturer='sbot_simulator',
                board_type=board_info.type_str,
                asset_tag=board_info.serial_number,
            )
            if (board := cls._get_valid_board(board_info.url, initial_identity)) is None:
                continue

            boards[board._identity.asset_tag] = board
        return MappingProxyType(boards)

    @classmethod
    def _get_supported_boards(
        cls, manual_boards: Optional[list[str]] = None,
    ) -> MappingProxyType[str, MotorBoard]:
        """
        Find all connected motor boards.

        Ports are filtered to the USB vendor and product ID: 0x0403 and 0x6001 respectively.

        :param manual_boards: A list of manually specified serial ports to also attempt
            to connect to, defaults to None
        :return: A mapping of serial numbers to motor boards.
        """
        if IN_SIMULATOR:
            return cls._get_simulator_boards()

        boards = {}
        serial_ports = comports()
        for port in serial_ports:
            # Filter to USB vendor and product ID of the FTDI FT232R
            # chip used on the motor board
            if port.vid == 0x0403 and port.pid == 0x6001:
                # Create board identity from USB port info
                initial_identity = get_USB_identity(port)

                if (board := cls._get_valid_board(port.device, initial_identity)) is None:
                    continue

                boards[board._identity.asset_tag] = board

        # Add any manually specified boards
        if isinstance(manual_boards, list):
            for manual_port in manual_boards:
                # Create board identity from the info given
                initial_identity = BoardIdentity(
                    board_type='manual',
                    asset_tag=manual_port,
                )

                if (board := cls._get_valid_board(manual_port, initial_identity)) is None:
                    continue

                boards[board._identity.asset_tag] = board
        return MappingProxyType(boards)

    @property
    @log_to_debug
    def motors(self) -> tuple[Motor, Motor]:
        """
        A tuple of the two motors on the board.

        :return: A tuple of the two motors on the board.
        """
        return self._motors

    @log_to_debug
    def identify(self) -> BoardIdentity:
        """
        Get the identity of the board.

        :return: The identity of the board.
        """
        response = self._serial.query('*IDN?')
        return BoardIdentity(*response.split(':'))

    @property
    @log_to_debug
    def status(self) -> MotorStatus:
        """
        The status of the board.

        :return: The status of the board.
        """
        response = self._serial.query('*STATUS?')
        return MotorStatus.from_status_response(response)

    @log_to_debug
    def reset(self) -> None:
        """
        Reset the board.

        This command disables the motors and clears all faults.
        """
        self._serial.write('*RESET')

    def _cleanup(self) -> None:
        """
        Disable the motors while exiting.

        This method is registered as an exit handler.
        """
        try:
            self.reset()
        except Exception:
            logger.warning(f"Failed to cleanup motor board {self._identity.asset_tag}.")

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial}>"


class Motor:
    """
    A class representing a motor on the motor board.

    Each motor is controlled through the power property
    and its current can be read using the current property.

    :param serial: The serial wrapper to use to communicate with the board.
    :param index: The index of the motor on the board.
    """
    __slots__ = ('_serial', '_index')

    def __init__(self, serial: SerialWrapper, index: int):
        self._serial = serial
        self._index = index

    @property
    @log_to_debug
    def power(self) -> float:
        """
        Read the current power setting of the motor.

        :return: The power of the motor as a float between -1.0 and 1.0
            or the special value MotorPower.COAST.
        """
        response = self._serial.query(f'MOT:{self._index}:GET?')

        data = response.split(':')
        enabled = (data[0] == '1')
        value = int(data[1])

        if not enabled:
            return MotorPower.COAST
        return map_to_float(value, -1000, 1000, -1.0, 1.0, precision=3)

    @power.setter
    @log_to_debug(setter=True)
    def power(self, value: float) -> None:
        """
        Set the power of the motor.

        Internally this method maps the power to an integer between
        -1000 and 1000 so only 3 digits of precision are available.

        :param value: The power of the motor as a float between -1.0 and 1.0
            or the special values MotorPower.COAST and MotorPower.BRAKE.
        """
        if value == MotorPower.COAST:
            self._serial.write(f'MOT:{self._index}:DISABLE')
            return
        value = float_bounds_check(
            value, -1.0, 1.0,
            'Motor power is a float between -1.0 and 1.0')

        setpoint = map_to_int(value, -1.0, 1.0, -1000, 1000)
        self._serial.write(f'MOT:{self._index}:SET:{setpoint}')

    @property
    @log_to_debug
    def current(self) -> float:
        """
        Read the current draw of the motor.

        :return: The current draw of the motor in amps.
        """
        response = self._serial.query(f'MOT:{self._index}:I?')
        return float(response) / 1000

    @property
    @log_to_debug
    def in_fault(self) -> bool:
        """
        Check if the motor is in a fault state.

        :return: True if the motor is in a fault state, False otherwise.
        """
        response = self._serial.query('*STATUS?')
        return MotorStatus.from_status_response(response).output_faults[self._index]

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__} index={self._index} {self._serial}>"


if __name__ == '__main__':  # pragma: no cover
    motorboards = MotorBoard._get_supported_boards()
    for serial_num, board in motorboards.items():
        print(serial_num)
        print(board.identify())



"""Functions for killing the robot after a certain amount of time."""
import atexit
import logging
import os
import signal
import sys
from threading import Timer
from types import FrameType
from typing import Optional

logger = logging.getLogger(__name__)

TIMEOUT_MESSAGE = "Timeout expired: Game Over!"
EXIT_ATTEMPTS = 0


def timeout_handler(signal_type: int, stack_frame: Optional[FrameType]) -> None:
    """
    Handle the timeout to kill the current process.

    This function is called when the timeout expires and will stop the robot's main process.
    In order for this to work, any threads that are created must be daemons.

    If the process doesn't terminate clearly (perhaps because the exception was caught),
    exit less cleanly.

    NOTE: This function is not called on Windows.

    :param signal_type: The signal that triggered this handler
    :param stack_frame: The stack frame at the time of the signal
    """
    global EXIT_ATTEMPTS
    EXIT_ATTEMPTS += 1

    if sys.platform == "win32":
        raise AssertionError("This function should not be called on Windows")

    if EXIT_ATTEMPTS == 1:
        # Allow 2 seconds for the process to exit cleanly before killing it
        signal.alarm(2)
        logger.info(TIMEOUT_MESSAGE)
        # Exit cleanly
        exit(0)
    else:
        # The process didn't exit cleanly, so first call the cleanup handlers
        # and use an unhanded alarm to force python to exit with a core dump
        signal.signal(signal.SIGALRM, signal.SIG_DFL)
        signal.alarm(2)  # Allow 2 seconds for cleanup

        atexit._run_exitfuncs()
        exit(0)


def win_timeout_handler() -> None:
    """
    Kill the main process on Windows.

    This function is called when the timeout expires and will stop the robot's main process.
    In order for this to work, any threads that are created must be daemons.

    NOTE: This function is only called on Windows.
    """
    logger.info(TIMEOUT_MESSAGE)
    os.kill(os.getpid(), signal.SIGTERM)


def kill_after_delay(timeout_seconds: int) -> None:
    """
    Kill the robot after a certain amount of time.

    Interrupts main process after the given delay.

    :param timeout_seconds: The number of seconds to wait before killing the robot
    """

    if sys.platform == "win32":
        # Windows doesn't have SIGALRM,
        # so we approximate its functionality using a delayed SIGTERM
        timer = Timer(timeout_seconds, win_timeout_handler)
        timer.daemon = True
        timer.start()
    else:
        signal.signal(signal.SIGALRM, timeout_handler)
        signal.alarm(timeout_seconds)
    logger.info(f"Timeout set: {timeout_seconds}s")



"""
Classes for marker detections and various axis representations.
"""
from math import atan2, hypot
from typing import NamedTuple, Tuple, cast

# import numpy as np
from april_vision import Marker as AprilMarker
from april_vision import Orientation as AprilOrientation
from numpy.typing import NDArray


class PixelCoordinates(NamedTuple):
    """
    Coordinates within an image made up from pixels.

    Floating point type is used to allow for subpixel detected locations
    to be represented.

    :param float x: X coordinate
    :param float y: Y coordinate
    """

    x: float
    y: float


class Coordinates(NamedTuple):
    """
    3D coordinates in space.

    :param float x: X coordinate
    :param float y: Y coordinate
    :param float z: Z coordinate
    """

    x: float
    y: float
    z: float


class Orientation(NamedTuple):
    """
    Orientation of a marker in space.

    :param yaw:   Yaw of the marker, a rotation about the vertical axis, in radians.
                  Positive values indicate a rotation clockwise from the perspective
                  of the marker.
                  Zero values have the marker facing the camera square-on.
    :param pitch: Pitch of the marker, a rotation about the transverse axis, in
                  radians.
                  Positive values indicate a rotation upwards from the perspective
                  of the marker.
                  Zero values have the marker facing the camera square-on.
    :param roll:  Roll of the marker, a rotation about the longitudinal axis,
                  in radians.
                  Positive values indicate a rotation clockwise from the perspective
                  of the marker.
                  Zero values have the marker facing the camera square-on.
    """

    yaw: float
    pitch: float
    roll: float


class Position(NamedTuple):
    """
    Position of a marker in space from the camera's perspective.

    :param distance:          Distance from the camera to the marker, in millimetres.
    :param horizontal_angle:  Horizontal angle from the camera to the marker, in radians.
                              Ranges from -pi to pi, with positive values indicating
                              markers to the right of the camera. Directly in front
                              of the camera is 0 rad.
    :param vertical_angle:    Vertical angle from the camera to the marker, in radians.
                              Ranges from -pi to pi, with positive values indicating
                              markers above the camera. Directly in front of the camera
                              is 0 rad.
    """

    distance: float
    horizontal_angle: float
    vertical_angle: float


PixelCorners = Tuple[PixelCoordinates, PixelCoordinates, PixelCoordinates, PixelCoordinates]


class Marker(NamedTuple):
    """
    Wrapper of a marker detection with axis and rotation calculated.

    :param id: The ID of the detected marker
    :param size: The physical size of the marker in millimeters
    :param pixel_corners: A tuple of the PixelCoordinates of the marker's corners in the frame
    :param pixel_centre: The PixelCoordinates of the marker's centre in the frame
    :param position: Position information of the marker relative to the camera
    :param orientation: Orientation information of the marker
    """

    id: int
    size: int
    pixel_corners: PixelCorners
    pixel_centre: PixelCoordinates

    position: Position = Position(0, 0, 0)
    orientation: Orientation = Orientation(0, 0, 0)

    @classmethod
    def from_april_vision_marker(cls, marker: AprilMarker) -> 'Marker':
        """Generate a marker object using the data from april_vision's Marker object."""
        if marker.rvec is None or marker.tvec is None:
            raise ValueError("Marker lacks pose information")

        _cartesian = cls._standardise_tvec(marker.tvec.flatten())
        _orientation = AprilOrientation.from_rvec_matrix(marker.rvec)

        return cls(
            id=marker.id,
            size=marker.size,
            pixel_corners=cast(
                PixelCorners,
                tuple(PixelCoordinates(*corner) for corner in marker.pixel_corners)),
            pixel_centre=PixelCoordinates(*marker.pixel_centre),

            position=Position(
                distance=int(hypot(*_cartesian) * 1000),
                horizontal_angle=atan2(-_cartesian.y, _cartesian.x),
                vertical_angle=atan2(_cartesian.z, _cartesian.x),
            ),

            orientation=Orientation(
                yaw=_orientation.yaw,
                pitch=_orientation.pitch,
                roll=_orientation.roll,
            ),
        )

    @staticmethod
    def _standardise_tvec(tvec: NDArray) -> Coordinates:
        """
        Standardise the tvec to use the marker's coordinate system.

        The marker's coordinate system is defined as:
        - X axis is straight out of the camera
        - Y axis is to the left of the camera
        - Z axis is up
        """
        return Coordinates(tvec[2], -tvec[0], -tvec[1])

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} id={self.id} distance={self.position.distance:.0f}mm "
            f"horizontal_angle={self.position.horizontal_angle:.2f}rad "
            f"vertical_angle={self.position.vertical_angle:.2f}rad size={self.size}mm>"
        )



import logging

import pytest

from sr.robot3.exceptions import BoardDisconnectionError
from sr.robot3.serial_wrapper import SerialWrapper, retry
from sr.robot3.utils import BoardIdentity


def test_retry_exception() -> None:
    """Test the retry decorator when an exception is repeatedly raised."""
    call_count = 0

    @retry(times=3, exceptions=Exception)
    def test_func() -> None:
        """Test function."""
        nonlocal call_count
        call_count += 1
        raise Exception("Test exception")

    with pytest.raises(Exception):
        test_func()

    # 4 calls for 3 retries
    assert call_count == 4


def test_retry_no_exception() -> None:
    """Test the retry decorator when no exception is raised."""
    call_count = 0

    @retry(times=3, exceptions=Exception)
    def test_func() -> None:
        """Test function."""
        nonlocal call_count
        call_count += 1
        return 'Test return'

    assert test_func() == 'Test return'

    # 1 call for no retries
    assert call_count == 1


def test_retry_single_exception() -> None:
    """Test the retry decorator."""
    call_count = 0

    @retry(times=3, exceptions=Exception)
    def test_func() -> None:
        """Test function."""
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            # Only throw an execption on the first call
            raise Exception("Test exception")
        return 'Test return'

    assert test_func() == 'Test return'

    # only first call raises an exception
    assert call_count == 2


def test_serial_wrapper(caplog) -> None:
    """
    Test the serial wrapper.

    Using a loopback serial port causes all data to be sent back to the sender.
    """
    # Capture all logging
    caplog.set_level(0)
    serial_wrapper = SerialWrapper(
        port='loop://',
        baud=115200,
        identity=BoardIdentity(
            board_type='Test board',
            sw_version='1.0',
            asset_tag='1234',
        ),
    )

    # Test that the serial port uses the correct identity
    assert str(serial_wrapper) == "<SerialWrapper 'loop://' '1234'>"

    # Test that we can set a new identity
    serial_wrapper.identity = BoardIdentity(
        board_type='Test board',
        sw_version='1.0',
        asset_tag='5678',
    )
    assert str(serial_wrapper) == "<SerialWrapper 'loop://' '5678'>"

    # Test that the serial port is opened on the first message
    assert not serial_wrapper.serial.is_open
    assert serial_wrapper.query("Echo test") == "Echo test"
    assert serial_wrapper.serial.is_open
    assert caplog.record_tuples == [
        ('sr.robot3.serial_wrapper', logging.INFO, 'Connected to board Test board: 5678'),
        ('sr.robot3.serial_wrapper', 5, "Serial write - 'Echo test'"),
        ('sr.robot3.serial_wrapper', 5, "Serial read  - 'Echo test'"),
    ]

    # Test that an exception is raised if a NACK is received
    with pytest.raises(RuntimeError, match="Test exception"):
        serial_wrapper.write("NACK:Test exception")


def test_serial_wrapper_invalid_port() -> None:
    """
    Test that a BoardDisconnectionError is raised when the serial port is cannot be opened.
    """
    serial_wrapper = SerialWrapper(
        port='socket://localhost:1',  # A port that has nothing listening on it
        baud=115200,
    )
    with pytest.raises(
        BoardDisconnectionError,
        match="Connection to board : could not be established"
    ):
        serial_wrapper.query("Echo test")


def test_serial_wrapper_message_timeout(caplog, monkeypatch) -> None:
    """
    Test that a BoardDisconnectionError is raised when the serial response times out.
    """
    # Capture all logging
    caplog.set_level(0)
    serial_wrapper = SerialWrapper(
        port='loop://',
        baud=115200,
    )

    # Allow the serial port to be opened
    assert serial_wrapper.query("Echo test") == "Echo test"
    caplog.clear()

    monkeypatch.setattr(serial_wrapper.serial, 'readline', lambda: b'')
    with pytest.raises(
        BoardDisconnectionError,
        match="Board : disconnected during transaction"
    ):
        serial_wrapper.query("Echo test")

    assert caplog.record_tuples == [
        ('sr.robot3.serial_wrapper', 5, "Serial write - 'Echo test'"),
        ('sr.robot3.serial_wrapper', 5, "Serial read  - ''"),
        ('sr.robot3.serial_wrapper', logging.WARNING,
         'Connection to board : timed out waiting for response'),
        ('sr.robot3.serial_wrapper', logging.WARNING, 'Board : disconnected'),

        ('sr.robot3.serial_wrapper', logging.INFO, 'Connected to board : '),
        ('sr.robot3.serial_wrapper', 5, "Serial write - 'Echo test'"),
        ('sr.robot3.serial_wrapper', 5, "Serial read  - ''"),
        ('sr.robot3.serial_wrapper', logging.WARNING,
         'Connection to board : timed out waiting for response'),
        ('sr.robot3.serial_wrapper', logging.WARNING, 'Board : disconnected'),

        ('sr.robot3.serial_wrapper', logging.INFO, 'Connected to board : '),
        ('sr.robot3.serial_wrapper', 5, "Serial write - 'Echo test'"),
        ('sr.robot3.serial_wrapper', 5, "Serial read  - ''"),
        ('sr.robot3.serial_wrapper', logging.WARNING,
         'Connection to board : timed out waiting for response'),
        ('sr.robot3.serial_wrapper', logging.WARNING, 'Board : disconnected'),

        ('sr.robot3.serial_wrapper', logging.INFO, 'Connected to board : '),
        ('sr.robot3.serial_wrapper', 5, "Serial write - 'Echo test'"),
        ('sr.robot3.serial_wrapper', 5, "Serial read  - ''"),
        ('sr.robot3.serial_wrapper', logging.WARNING,
         'Connection to board : timed out waiting for response'),
        ('sr.robot3.serial_wrapper', logging.WARNING, 'Board : disconnected'),
    ]



"""Test the timeout function."""
from time import sleep, time

import os
from pathlib import Path
import signal
import subprocess
import sys
import pytest
from unittest.mock import Mock

from sr.robot3.timeout import kill_after_delay

TEST_FILES = list((Path(__file__).parent / 'test_data/timeout_scripts').iterdir())
EXTRA_TEST_FILES_DIR = Path(__file__).parent / 'test_data/timeout_scripts_extra'

@pytest.mark.skipif(sys.platform == "win32", reason="does not run on Windows")
def test_kill_after_delay() -> None:
    """Test that the process is killed within the time."""
    with pytest.raises(SystemExit):
        kill_after_delay(2)
        sleep(3)

    # Clear the set alarm
    signal.alarm(0)


@pytest.mark.skipif(sys.platform != "win32", reason="only runs on Windows")
def test_kill_after_delay_windows(monkeypatch) -> None:
    """Test that the process is killed within the time on windows."""
    kill = Mock()
    pid = os.getpid()
    monkeypatch.setattr(os, "kill", kill)

    kill_after_delay(2)
    sleep(1)

    kill.assert_not_called()

    sleep(1.5)
    kill.assert_called_once_with(pid, signal.SIGTERM)


@pytest.mark.parametrize(
    "test_file",
    TEST_FILES,
    ids=[f.stem for f in TEST_FILES]
)
def test_kill_after_delay_e2e(test_file: Path) -> None:
    start_time = time()
    child = subprocess.Popen([
        sys.executable,
        str(test_file),
    ], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    child.wait(timeout=8)
    run_time = time() - start_time

    assert 2 < run_time < 8

    if sys.platform == "win32":
        # Windows terminates uncleanly
        assert child.returncode == signal.SIGTERM
    elif run_time < 4:
        # If the process terminated quickly, it should be successful
        assert child.returncode == 0
    else:
        # If the process took too long, it was killed ungracefully
        assert child.returncode == -signal.SIGALRM

def test_early_exit() -> None:
    start_time = time()
    child = subprocess.Popen([
        sys.executable,
        str(EXTRA_TEST_FILES_DIR / "early-exit.py"),
    ], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    child.wait(timeout=6)

    run_time = time() - start_time

    assert run_time < 2

    assert child.returncode == 0

def test_exception() -> None:
    start_time = time()
    child = subprocess.Popen([
        sys.executable,
        str(EXTRA_TEST_FILES_DIR / "exception.py"),
    ], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    child.wait(timeout=6)

    run_time = time() - start_time

    assert run_time < 2

    assert child.returncode == 1



"""Test log_to_debug functions."""
import logging

from sr.robot3.logging import log_to_debug


def test_log_to_debug(caplog):
    """Test log_to_debug."""
    @log_to_debug
    def func(a, b, c):
        return a + b + c

    with caplog.at_level("DEBUG"):
        func(1, 2, c=3)

    assert caplog.record_tuples == [
        (
            "tests.test_logging", logging.DEBUG,
            "Calling test_log_to_debug.<locals>.func(1, 2, c=3)"
        ),
        (
            "tests.test_logging", logging.DEBUG,
            "'test_log_to_debug.<locals>.func' returned 6"
        ),
    ]


def test_log_to_debug_setter(caplog):
    """Test log_to_debug with setter."""
    @log_to_debug(setter=True)
    def func(a, b, c):
        return a + b + c

    with caplog.at_level("DEBUG"):
        func(1, 2, c=3)

    assert caplog.record_tuples == [
        (
            "tests.test_logging", logging.DEBUG,
            "Calling setter test_log_to_debug_setter.<locals>.func(1, 2, c=3)"
        ),
    ]



"""The main entry point for the sr-robot3 package."""
from __future__ import annotations

import itertools
import logging
import os
import time
from pathlib import Path
from socket import socket
from types import MappingProxyType
from typing import Mapping, Optional, final

from . import game_specific, timeout
from ._version import __version__
from .arduino import Arduino
from .astoria import Metadata, RobotMode, init_astoria_mqtt
from .camera import AprilCamera, _setup_cameras
from .exceptions import MetadataNotReadyError
from .kch import KCH
from .logging import log_to_debug, setup_logging
from .motor_board import MotorBoard
from .power_board import Note, PowerBoard
from .raw_serial import RawSerial
from .servo_board import ServoBoard
from .simulator.time_server import TimeServer
from .utils import (
    IN_SIMULATOR, FinalMeta, ensure_atexit_on_term, obtain_lock, singular,
)

logger = logging.getLogger(__name__)


@final
class Robot(metaclass=FinalMeta):
    """
    The main robot class that provides access to all the boards.

    There can be only one instance of this class active in your operating
    system at a time, creating a second instance will raise an OSError.

    :param debug: Enable debug logging to the console, defaults to False
    :param wait_for_start: Wait in the constructor until the start button is pressed,
        defaults to True
    :param trace_logging: Enable trace level logging to the console, defaults to False
    :param ignored_arduinos: A list of Arduino serial numbers to avoid connecting to
    :param manual_boards: A dictionary of board types to a list of serial port paths
        to allow for connecting to boards that are not automatically detected, defaults to None
    :param raw_ports: A list of serial number, baudrate tuples to try connecting to.
    :param no_powerboard: If True, initialize the robot without a powerboard, defaults to False
    """
    __slots__ = (
        '_lock', '_metadata', '_power_board', '_motor_boards', '_servo_boards',
        '_arduinos', '_cameras', '_mqtt', '_astoria', '_kch', '_raw_ports',
        '_time_server', '_no_powerboard',
    )

    def __init__(
        self,
        *,
        debug: bool = False,
        wait_for_start: bool = True,
        trace_logging: bool = False,
        ignored_arduinos: Optional[list[str]] = None,
        manual_boards: Optional[dict[str, list[str]]] = None,
        raw_ports: Optional[list[tuple[str, int]]] = None,
        no_powerboard: bool = False,
    ) -> None:
        self._lock: TimeServer | socket | None
        if IN_SIMULATOR:
            self._lock = TimeServer.initialise()
            if self._lock is None:
                raise OSError(
                    'Unable to obtain lock, Is another robot instance already running?'
                )
        else:
            self._lock = obtain_lock()
        self._metadata: Optional[Metadata] = None
        self._no_powerboard = no_powerboard

        setup_logging(debug, trace_logging)
        ensure_atexit_on_term()

        logger.info(f"sr.robot3 version {__version__}")

        if IN_SIMULATOR:
            self._mqtt = None
            self._astoria = None
        else:
            self._mqtt, self._astoria = init_astoria_mqtt()

        if manual_boards:
            self._init_power_board(manual_boards.get(PowerBoard.get_board_type(), []))
        else:
            self._init_power_board()
        self._init_aux_boards(manual_boards, ignored_arduinos, raw_ports)
        self._init_camera()
        self._log_connected_boards()

        if wait_for_start:
            logger.debug("Waiting for start button.")
            self.wait_start()
        else:
            logger.debug("Not waiting for start button.")

    def _init_power_board(self, manual_boards: Optional[list[str]] = None) -> None:
        """
        Locate the PowerBoard and enable all the outputs to power the other boards.

        :param manual_boards: Serial port paths to also check for power boards,
            defaults to None
        :raises RuntimeError: If exactly one PowerBoard is not found
        """
        if self._no_powerboard:
            return

        power_boards = PowerBoard._get_supported_boards(manual_boards, sleep_fn=self.sleep)
        self._power_board = singular(power_boards)

        # Enable all the outputs, so that we can find other boards.
        self._power_board.outputs.power_on()

    def _init_aux_boards(
        self,
        manual_boards: Optional[dict[str, list[str]]] = None,
        ignored_arduinos: Optional[list[str]] = None,
        raw_ports: Optional[list[tuple[str, int]]] = None,
    ) -> None:
        """
        Locate the motor boards, servo boards, and Arduinos.

        All boards are located automatically, but additional serial ports can be
        provided using the manual_boards parameter. Located boards are queried for
        their identity and firmware version.

        :param manual_boards:  A dictionary of board types to a list of additional
            serial port paths that should be checked for boards of that type, defaults to None
        :param ignored_arduinos: A list of Arduino serial numbers to avoid connecting to
        :param raw_ports: A list of serial number, baudrate tuples to try connecting to
        """
        self._raw_ports: MappingProxyType[str, RawSerial] = MappingProxyType({})
        if manual_boards is None:
            manual_boards = {}

        manual_motorboards = manual_boards.get(MotorBoard.get_board_type(), [])
        manual_servoboards = manual_boards.get(ServoBoard.get_board_type(), [])
        manual_arduinos = manual_boards.get(Arduino.get_board_type(), [])

        self._kch = KCH()
        self._motor_boards = MotorBoard._get_supported_boards(manual_motorboards)
        self._servo_boards = ServoBoard._get_supported_boards(manual_servoboards)
        self._arduinos = Arduino._get_supported_boards(manual_arduinos, ignored_arduinos)
        if raw_ports:
            if not IN_SIMULATOR:
                self._raw_ports = RawSerial._get_supported_boards(raw_ports)
            else:
                logger.warning("Raw ports are not available in the simulator.")

    def _init_camera(self) -> None:
        """
        Locate cameras that we have calibration data for.

        These cameras are used for AprilTag detection and provide location data of
        markers in its field of view.
        """
        if IN_SIMULATOR:
            publish_fn = None
        else:
            assert self._mqtt is not None
            publish_fn = self._mqtt.wrapped_publish

        self._cameras = MappingProxyType(_setup_cameras(
            game_specific.MARKER_SIZES,
            publish_fn,
        ))

    def _log_connected_boards(self) -> None:
        """
        Log the board types and serial numbers of all the boards connected to the robot.

        Firmware versions are also logged at debug level.
        """
        # we only have one power board so make it iterable
        power_board = [] if self._no_powerboard else [self.power_board]
        boards = itertools.chain(
            power_board,
            self.motor_boards.values(),
            self.servo_boards.values(),
            self.arduinos.values(),
            self._cameras.values(),
            self.raw_serial_devices.values(),
        )
        for board in boards:
            identity = board.identify()
            board_type = board.__class__.__name__
            logger.info(f"Found {board_type}, serial: {identity.asset_tag}")
            logger.debug(
                f"Firmware Version of {identity.asset_tag}: {identity.sw_version}, "
                f"reported type: {identity.board_type}",
            )

    @property
    def kch(self) -> KCH:
        """
        Access the Raspberry Pi hat, including user-accessible LEDs.

        :returns: The KCH object
        """
        return self._kch

    @property
    def power_board(self) -> PowerBoard:
        """
        Access the power board connected to the robot.

        :return: The power board object
        """
        if self._no_powerboard:
            raise RuntimeError("No power board was initialized")

        return self._power_board

    @property
    def motor_boards(self) -> Mapping[str, MotorBoard]:
        """
        Access the motor boards connected to the robot.

        These are indexed by their serial number.

        :return: A mapping of serial numbers to motor boards
        """
        return self._motor_boards

    @property
    def motor_board(self) -> MotorBoard:
        """
        Access the motor board connected to the robot.

        This can only be used if there is exactly one motor board connected.

        :return: The motor board object
        :raises RuntimeError: If there is not exactly one motor board connected
        """
        return singular(self._motor_boards)

    @property
    def servo_boards(self) -> Mapping[str, ServoBoard]:
        """
        Access the servo boards connected to the robot.

        These are indexed by their serial number.

        :return: A mapping of serial numbers to servo boards
        """
        return self._servo_boards

    @property
    def servo_board(self) -> ServoBoard:
        """
        Access the servo board connected to the robot.

        This can only be used if there is exactly one servo board connected.

        :return: The servo board object
        :raises RuntimeError: If there is not exactly one servo board connected
        """
        return singular(self._servo_boards)

    @property
    def arduinos(self) -> Mapping[str, Arduino]:
        """
        Access the Arduinos connected to the robot.

        These are indexed by their serial number.

        :return: A mapping of serial numbers to Arduinos
        """
        return self._arduinos

    @property
    def arduino(self) -> Arduino:
        """
        Access the Arduino connected to the robot.

        This can only be used if there is exactly one Arduino connected.

        :return: The Arduino object
        :raises RuntimeError: If there is not exactly one Arduino connected
        """
        return singular(self._arduinos)

    @property
    def camera(self) -> AprilCamera:
        """
        Access the camera connected to the robot.

        This can only be used if there is exactly one camera connected.
        The robot class currently only supports one camera.

        :return: The camera object
        :raises RuntimeError: If there is not exactly one camera connected
        """
        return singular(self._cameras)

    @property
    def raw_serial_devices(self) -> Mapping[str, RawSerial]:
        """
        Access the raw serial devices connected to the robot.

        These are populated by the `raw_ports` parameter of the constructor,
        and are indexed by their serial number.

        :return: A mapping of serial numbers to raw serial devices
        """
        return self._raw_ports

    @log_to_debug
    def sleep(self, secs: float) -> None:
        """
        Sleep for a number of seconds.

        This is a convenience method that should be used instead of time.sleep()
        to make your code compatible with the simulator.

        :param secs: The number of seconds to sleep for
        """
        if IN_SIMULATOR:
            assert isinstance(self._lock, TimeServer)
            self._lock.sleep(secs)
        else:
            time.sleep(secs)

    @log_to_debug
    def time(self) -> float:
        """
        Get the number of seconds since the Unix Epoch.

        This is a convenience method that should be used instead of time.time()
        to make your code compatible with the simulator.

        NOTE: The robot's clock resets each time the robot is restarted, so this
        will not be the correct actual time but can be used to measure elapsed time.

        :returns: the number of seconds since the Unix Epoch.
        """
        if IN_SIMULATOR:
            assert isinstance(self._lock, TimeServer)
            return self._lock.get_time()
        else:
            return time.time()

    @property
    @log_to_debug
    def arena(self) -> str:
        """
        Get the arena that the robot is in.

        :return: The robot's arena
        :raises MetadataNotReadyError: If the start button has not been pressed yet
        """
        if self._metadata is None:
            raise MetadataNotReadyError()
        else:
            return self._metadata.arena

    @property
    @log_to_debug
    def zone(self) -> int:
        """
        Get the zone that the robot is in.

        :return: The robot's zone number
        :raises MetadataNotReadyError: If the start button has not been pressed yet
        """
        if self._metadata is None:
            raise MetadataNotReadyError()
        else:
            return self._metadata.zone

    @property
    @log_to_debug
    def mode(self) -> RobotMode:
        """
        Get the mode that the robot is in.

        :return: The robot's current mode
        :raises MetadataNotReadyError: If the start button has not been pressed yet
        """
        if self._metadata is None:
            raise MetadataNotReadyError()
        else:
            return self._metadata.mode

    @property
    def usbkey(self) -> Path:
        """
        The path of the USB code drive.

        :returns: path to the mountpoint of the USB code drive.
        """
        if IN_SIMULATOR:
            return Path(os.environ['SBOT_USBKEY_PATH'])
        else:
            assert self._astoria is not None
            return self._astoria.fetch_mount_path()

    @property
    def is_simulated(self) -> bool:
        """
        Determine whether the robot is simulated.

        :returns: True if the robot is simulated. False otherwise.
        """
        return IN_SIMULATOR

    @log_to_debug
    def wait_start(self) -> None:
        """
        Wait for the start button to be pressed.

        The power board will beep once when waiting for the start button.
        The power board's run LED will flash while waiting for the start button.
        Once the start button is pressed, the metadata will be loaded and the timeout
        will start if in competition mode.
        """
        def null_button_pressed() -> bool:
            return False

        if self._no_powerboard:
            # null out the start button function
            start_button_pressed = null_button_pressed
        else:
            start_button_pressed = self.power_board._start_button

        if IN_SIMULATOR:
            remote_start_pressed = null_button_pressed
        else:
            assert self._astoria is not None
            remote_start_pressed = self._astoria.get_start_button_pressed

        # ignore previous button presses
        _ = start_button_pressed()
        _ = remote_start_pressed()
        logger.info('Waiting for start button.')

        if not self._no_powerboard:
            self.power_board.piezo.buzz(Note.A6, 0.1)
            self.power_board._run_led.flash()
        self.kch._flash_start()

        while not (start_button_pressed() or remote_start_pressed()):
            self.sleep(0.1)
        logger.info("Start signal received; continuing.")
        if not self._no_powerboard:
            self.power_board._run_led.on()
        self.kch._start = False

        if IN_SIMULATOR:
            self._metadata = Metadata.parse_file(
                Path(os.environ['SBOT_METADATA_PATH']) / 'astoria.json'
            )
        else:
            assert self._astoria is not None
            # Load the latest metadata that we have received over MQTT
            self._metadata = self._astoria.fetch_current_metadata()

        # Simulator timeout is handled by the simulator supervisor
        if self._metadata.game_timeout and not IN_SIMULATOR:
            timeout.kill_after_delay(self._metadata.game_timeout)



"""
Test that raw serial devices can be created and used.

This test uses a mock serial wrapper to simulate the connection to the serial device.
"""
from __future__ import annotations

from typing import NamedTuple
from unittest.mock import MagicMock, call, patch

import pytest

from sr.robot3.raw_serial import RawSerial, RawSerialDevice
from sr.robot3.utils import BoardIdentity


class MockRawSerial(NamedTuple):
    """A mock arduino board."""

    serial_wrapper: MagicMock
    serial_device: RawSerial


@pytest.fixture
def raw_serial() -> None:
    with patch('sr.robot3.raw_serial.serial_for_url') as MockSerial:
        MockSerial.return_value = MagicMock()
        raw_serial_device = RawSerial('test://', identity=BoardIdentity(asset_tag='TEST123'))
        assert MockSerial.call_count == 1

        yield MockRawSerial(MockSerial, raw_serial_device)


def test_invalid_properties(raw_serial: MockRawSerial) -> None:
    """
    Test that settng invalid properties raise an AttributeError.
    """
    serial_device = raw_serial.serial_device

    with pytest.raises(AttributeError):
        serial_device.invalid_property = 1

    assert raw_serial.serial_wrapper.mock_calls == [
        call('test://', baudrate=115200, timeout=0.5, do_not_open=True),
        call().open(),
        call().reset_input_buffer(),
    ]
    assert raw_serial.serial_wrapper.return_value.mock_calls == [
        call.open(),
        call.reset_input_buffer(),
    ]


def test_serial_device_discovery(monkeypatch) -> None:
    """
    Test that discovery finds arduino boards from USB serial ports.

    Test that different USB pid/vid combinations are ignored.
    """
    class ListPortInfo(NamedTuple):
        """A mock serial port info."""
        device: str
        manufacturer: str
        product: str
        serial_number: str
        vid: int
        pid: int

    def mock_comports() -> list[ListPortInfo]:
        ports = [
            ListPortInfo(
                device='test://1',
                manufacturer='Student Robotics',
                product='Arduino',
                serial_number='TEST123',
                vid=0x2341,
                pid=0x0043,
            ),
            ListPortInfo(  # A arduino board with a different pid/vid
                device='test://3',
                manufacturer='Other',
                product='Arduino',
                serial_number='OTHER',
                vid=0x1234,
                pid=0x5678,
            ),
            ListPortInfo(  # An unrelated device
                device='test://5',
                manufacturer='Student Robotics',
                product='OTHER',
                serial_number='TESTABC',
                vid=0x2341,
                pid=0x0043,
            ),
        ]
        return ports

    monkeypatch.setattr('sr.robot3.raw_serial.comports', mock_comports)
    with patch('sr.robot3.raw_serial.serial_for_url') as MockSerial:
        MockSerial.return_value = MagicMock()
        serial_devices = RawSerial._get_supported_boards([
            RawSerialDevice(
                serial_number='TEST123',
            ),
            RawSerialDevice(
                serial_number='OTHER',
                baudrate=9600,
            ),
        ])
        assert MockSerial.mock_calls == [
            call('test://1', baudrate=115200, timeout=0.5, do_not_open=True),
            call().open(),
            call().reset_input_buffer(),
            call('test://3', baudrate=9600, timeout=0.5, do_not_open=True),
            call().open(),
            call().reset_input_buffer(),
        ]

        assert len(serial_devices) == 2
        assert {'TEST123', 'OTHER'} == set(serial_devices.keys())

        assert serial_devices['TEST123'].identify() == BoardIdentity(
            manufacturer='Student Robotics',
            board_type='Arduino',
            asset_tag='TEST123',
            sw_version='2341:0043',
        )
        assert serial_devices['OTHER'].identify() == BoardIdentity(
            manufacturer='Other',
            board_type='Arduino',
            asset_tag='OTHER',
            sw_version='1234:5678',
        )


def test_serial_device(raw_serial: MockRawSerial) -> None:
    serial_device = raw_serial.serial_device
    serial_object = raw_serial.serial_wrapper.return_value

    serial_device.write(b'hello')
    assert serial_object.write.mock_calls == [call(b'hello')]

    serial_object.read.return_value = b'A'

    assert serial_device.read(1) == b'A'
    assert serial_object.read.mock_calls == [call(1)]

    serial_object.read_until.return_value = b'hello'

    assert serial_device.read_until(b'x') == b'hello'
    assert serial_object.read_until.mock_calls == [call(b'x')]
